@@ -1,10 +1,13 @@
-"""A fast channel state information parser for Intel, Atheros, Nexmon and
-ESP32."""
+"""A fast channel state information parser for Intel, Atheros, Nexmon, ESP32
+and Picoscenes."""
 
 import os
-import numpy as np
 
+import numpy as np
+from numpy.lib import recfunctions as rfn
 from . import _csiread
+from . import _picoscenes
+from ._type import init_dtype_picoscenes
 
 
 def stringify(array, sep=':'):
@@ -261,7 +264,7 @@ class Intel(_csiread.Intel):
         return super().apply_sm(scaled_csi)
 
     def display(self, index):
-        """Prints the formatted representation of ``index`` packet"""
+        """Print the formatted representation of ``index`` packet"""
         T = "%s%-20s: %s\n"
         tab = " " * 2
 
@@ -469,7 +472,7 @@ class Atheros(_csiread.Atheros):
         return super().readstp(endian)
 
     def display(self, index):
-        """Prints the formatted representation of ``index`` packet"""
+        """Print the formatted representation of ``index`` packet"""
         T = "%s%-20s: %s\n"
         tab = " " * 2
 
@@ -657,7 +660,7 @@ class Nexmon(_csiread.Nexmon):
         return _nex_group(self.seq, self.core, self.spatial, s_num, c_num)
 
     def display(self, index):
-        """Prints the formatted representation of ``index`` packet"""
+        """Print the formatted representation of ``index`` packet"""
         T = "%s%-20s: %s\n"
         tab = " " * 2
 
@@ -709,14 +712,14 @@ class NexmonPull46(_csiread.NexmonPull46):
         See ``Nexmon``
 
     Attributes:
-        _autoscale (int): It can be 0 and 1, Pull 46 set it to 0. If you got a
-            bad plotting, you may want it to be 1.
+        _autoscale (int): It can be 0 and 1
         rssi (ndarray): rssi
         fc (ndarray): frame control
         others: see ``Nexmon``
 
     References:
         1. `nexmon_csi pull 46 <https://github.com/seemoo-lab/nexmon_csi/pull/46>`_
+        2. `nexmon_csi pull 256 <https://github.com/seemoo-lab/nexmon_csi/pull/256>`_
     """
     def __init__(self, file, chip, bw, if_report=True, bufsize=0):
         super(NexmonPull46, self).__init__(file, chip, bw, if_report, bufsize)
@@ -759,7 +762,7 @@ class NexmonPull46(_csiread.NexmonPull46):
         return _nex_group(self.seq, self.core, self.spatial, s_num, c_num)
 
     def display(self, index):
-        """Prints the formatted representation of ``index`` packet"""
+        """Print the formatted representation of ``index`` packet"""
         T = "%s%-20s: %s\n"
         tab = " " * 2
 
@@ -784,6 +787,9 @@ class NexmonPull46(_csiread.NexmonPull46):
         print(s, end='')
 
 
+NexmonPull256 = NexmonPull46
+
+
 class ESP32:
     """Parse CSI obtained using 'ESP32-CSI-Tool'.(experimental)
 
@@ -957,7 +963,7 @@ class ESP32:
         self.count = count
 
     def display(self, index):
-        """Prints the formatted representation of ``index`` packet"""
+        """Print the formatted representation of ``index`` packet"""
         T = "%s%-20s: %s\n"
         tab = " " * 2
 
@@ -992,3 +998,313 @@ def _nex_group(seq, core, spatial, c_num=4, s_num=4):
     offset = offset.reshape(-1, c_num, s_num)
 
     return offset
+
+
+class Picoscenes(_picoscenes.Picoscenes):
+    """Parse CSI obtained using 'PicoScenes'.
+
+    Args:
+        file (str or None): CSI data file ``.csi``. If ``str``, ``read``
+            methods is allowed. If ``None``, ``seek`` and ``pmsg`` methods are
+            allowed.
+        pl_size (dict, optional): `pl_size` will set the shape of some fields
+            in structured array ``raw``. These fields store the attributes
+            which are variable length arrays. These attributes of one frame
+            will only be parsed, when structured array is large enough to store
+            them. Otherwise they will be skipped. The ``check`` method can help
+            you set ``pl_size``. Default: ``None``.
+        if_report (bool, optional): Report the parsed result. Default: `True`
+        bufsize (int, optional): The maximum amount of frames to be parsed. If
+            ``0`` and file is ``str``, all frames will be parsed. If ``0`` and
+            file is ``None``, this parameter is ignored by `pmsg` method.
+            Default: 0
+
+    Attributes:
+        file (str, readonly): CSI data file
+        count (int, readonly): Count of csi frames parsed
+        pl_size (dict): A dictionary which initializes the dtype of ``raw``
+        raw (ndarray): structured array which stores the parsed result, See
+            ``PicoScenes documentation: PicoScenes MATLAB Toolbox`` for more
+            details.
+        dynamic attributes (recarray): Attribute names are ``raw.dtype.names``. 
+            They are just the aliases of ``raw``. Only exist after calling
+            ``read`` method.
+
+    Note:
+        1. Edge padding (Pads with the edge values of array) are applied to
+        raw["xxx"]["SubcarrierIndex"]
+
+    Examples:
+
+        >>> pl_size_rx_by_iwl5300 = {
+        >>>     'CSI': [30, 3, 2],
+        >>>     'PilotCSI': [0, 0, 0],
+        >>>     'LegacyCSI': [0, 0, 0],
+        >>>     'BasebandSignals': [0, 0, 0],
+        >>>     'PreEQSymbols': [0, 0, 0],
+        >>>     'MPDU': 1522
+        >>> }
+        >>> csifile = "../material/picoscenes/dataset/rx_by_iwl5300.csi"
+        >>> csidata = csiread.Picoscenes(csifile, pl_size_rx_by_iwl5300)
+        >>> csidata.read()
+        >>> csidata.check()
+        >>> csidata.display(2)
+        >>> print(csidata.raw['CSI']['CSI'][2].shape)
+        >>> print(csidata.CSI.CSI.shape)
+
+    References:
+        1. Zhiping Jiang, Tom H. Luan, Xincheng Ren, Dongtao Lv, Han Hao,
+        Jing Wang, Kun Zhao, Wei Xi, Yueshen Xu, Rui Li, Eliminating the
+        Barriers: Demystifying Wi-Fi Baseband Design and Introducing the
+        PicoScenes Wi-Fi Sensing Platform, in IEEE Internet of Things
+        Journal (IEEE IOT-J), doi: 10.1109/JIOT.2021.3104666, preprint
+        on arxiv.
+        2. `PicoScenes documentation <https://ps.zpj.io>`_
+    """
+    def __init__(self, file, pl_size=None, if_report=True, bufsize=0):
+        self.pl_size = self.__init_pl_size(pl_size)
+        dtype = init_dtype_picoscenes(self.pl_size)
+        super(Picoscenes, self).__init__(file, dtype, if_report, bufsize)
+
+    def __getitem__(self, index):
+        return self.raw[index]
+
+    def __init_pl_size(self, pl_size):
+        ret = {
+            'CSI': [0, 0, 0],
+            'PilotCSI': [0, 0, 0],
+            'LegacyCSI': [0, 0, 0],
+            'BasebandSignals': [0, 0, 0],
+            'PreEQSymbols': [0, 0, 0],
+            'MPDU': 0
+        }
+        if isinstance(pl_size, dict):
+            ret.update(pl_size)
+        elif isinstance(pl_size, list):
+            assert len(pl_size) == len(ret)
+            ret = {k: v for k, v in zip(ret.keys(), pl_size)}
+        elif isinstance(pl_size, tuple()):
+            assert len(pl_size) == len(ret)
+            ret = {k: v for k, v in zip(ret.keys(), pl_size)}
+        else:
+            pass
+        return ret
+
+    def _merge(self):
+        """Set dynamic attributes
+
+        Note: 
+            1. Don't run this method in a loop, it will slow down the program.
+            2. ndim, shape and itemsize of `info` conflict with ndarray's
+                attributes. e.g. PreEQSymbols.info.ndim will be 0, not 3.
+        """
+        for name in self.raw.dtype.names:
+            self.__setattr__(name, self.raw[name].view(np.recarray))
+
+    def read(self):
+        """Parse data
+
+        Examples:
+
+            >>> csifile = "../material/picoscenes/dataset/rx_by_iwl5300.csi"
+            >>> csidata = csiread.Picoscenes(csifile, {"CSI": (30, 3, 2)})
+            >>> csidata.read()
+            >>> print(csidata.raw[10]["CSI"].dtype.names)
+
+        """
+        super().read()
+        self._merge()
+
+    def seek(self, file, pos, num):
+        """Read frames from a specific position
+
+        This method allows us to read different parts of different files
+        randomly. It could be useful in Machine Learning. However, it could be
+        very slow when reading files in HDD for the first time. For this case,
+        it is better to do a pre-read with ``read()`` first.
+
+        Args:
+            file (str): CSI data file.
+            pos (int): Position of file descriptor corresponding to the frame.
+                Currently, it must be returned by the function in
+                ``example/csiseek.py``.
+            num (int): Number of frames to be read. If ``0``, all frames
+                after ``pos`` will be read.
+
+        Examples:
+
+            >>> csifile = "../material/picoscenes/dataset/rx_by_iwl5300.csi"
+            >>> csidata = csiread.Picoscenes(csifile, {"CSI": (30, 3, 2)})
+            >>> for i in range(10):
+            >>>     csidata.seek(csifile, 0, i+1)
+            >>>     print(csidata.raw["CSI"]["CSI"].shape)
+        """
+        super().seek(file, pos, num)
+
+    def pmsg(self, data):
+        """Parse message in real time (This method hasn't been READY)
+
+        Args:
+            data (bytes): A bytes object representing the data received by udp
+                socket
+        Returns:
+            int: The status code. If ``0xf300``, parse message successfully.
+                Otherwise, the ``data`` is not a PicoScenes frame.
+
+        Examples:
+
+            >>> import socket
+            >>> import csiread
+            >>>
+            >>> csidata = csiread.Picoscenes(None, {'CSI': [30, 3, 3]})
+            >>> with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
+            >>>     s.bind(('127.0.0.1', 10011))
+            >>>     while True:
+            >>>         data, address_src = s.recvfrom(4096)
+            >>>         code = csidata.pmsg(data)
+            >>>         if code == 0xf300:
+            >>>             print(csidata.raw[0]["CSI"]["CSI"].shape)
+        """
+        return super().pmsg(data)
+
+    def interpolate_csi(self, name='CSI', mode='AP'):
+        """Interpolate csi by linear method
+
+        Args:
+            name (str, optional): CSI type, It can be "CSI" and "LegacyCSI",
+                Default: "CSI"
+            mode (str, optional): How to interpolate csi. It can be "AP" and
+                "IQ", "AP" means interpolating along amplitude and phase; "IQ"
+                means interpolating along IQ signal. Default: "AP"
+
+        Returns:
+            (tuple)
+            interp_csi (ndarray): csi after interpolateing
+            interp_scindex (ndarray): subcarrier index after interpolateing
+        """
+        assert name in ["CSI", "LegacyCSI"]
+        assert mode in ["AP", "IQ"]
+        flag_IQ = True if mode == 'IQ' else False
+        interp_csi, interp_scindex = super().interpolate_csi(name, flag_IQ)
+        return interp_csi, interp_scindex
+
+    def _get_info_shape(self, name):
+        """return `name`'s shape
+
+        Args:
+            name (str): it can be "CSI", "PilotCSI", "LegacyCSI",
+                "BasebandSignals", "PreEQSymbols" and "MPDU"
+        Returns:
+            info (ndarray): 2D array which records the shape of `name`
+        """
+        if name in ["CSI", "PilotCSI", "LegacyCSI"]:
+            nsc = self.raw[name]["Info"]["NumTones"]
+            nrx = self.raw[name]["Info"]["NumRx"]
+            ntx = self.raw[name]["Info"]["NumTx"] \
+                + self.raw[name]["Info"]["NumESS"]
+            info = np.c_[nsc, nrx, ntx]
+        elif name in ["BasebandSignals", "PreEQSymbols"]:
+            info = self.raw[name]["Info"]["Shape"]
+        elif name in ["MPDU"]:
+            info = self.raw[name]["Info"]["Length"][:, None]
+        else:
+            info = None
+        return info
+
+    def check(self):
+        """helper method for setting parameter `pl_size`"""
+        T = "%15s%15s%15s%15s%10s\n"
+        Delimiter = T % ("-"*15, "-"*15, "-"*15, "-"*15, "-"*10)
+
+        def rowstring(name):
+            info = self._get_info_shape(name)
+            pl = np.asarray(self.pl_size[name])
+            skip = np.nan if info.max() == 0 else (info > pl).any(1).sum()
+            with np.printoptions(formatter={'all':lambda x: str(x)}):
+                s = T % (name,
+                         info.min(0) if pl.ndim else info.min(0)[0],
+                         info.max(0) if pl.ndim else info.max(0)[0],
+                         pl, skip)
+            return s
+
+        s = ""
+        s += Delimiter
+        s += T % ("CHECK", "min", "max", "pl_size", "skip")
+        s += Delimiter
+        s += rowstring("CSI")
+        s += rowstring("PilotCSI")
+        s += rowstring("LegacyCSI")
+        s += rowstring("BasebandSignals")
+        s += rowstring("PreEQSymbols")
+        s += rowstring("MPDU")
+        s += Delimiter
+        print(s, end='')
+
+    def display(self, index):
+        """Print the formatted representation of ``index`` frame"""
+        T = "%s%-20s: %s\n"
+
+        AtherosCFTuningPolicy = {
+            0: "0",                         # None
+            30: "30 (CFTuningByChansel)",
+            31: "31 (CFTuningByFastCC)",
+            32: "32 (CFTuningByHardwareReset)",
+            33: "33 (CFTuningByDefault)"
+        }
+
+        hfo = len(self.raw['RxExtraInfo'].dtype.names) // 2
+
+        def skip(raw, name):
+            if name == 'MVMExtra' and raw[name]['FTMClock'] == 0:
+                return True
+            if name == 'PicoScenesHeader' and raw[name]['MagicValue'] == 0:
+                return True
+            if name == 'TxExtraInfo' and raw[name]['Version'] == 0:
+                return True
+            if name == 'PilotCSI' and raw[name]['Info']['DeviceType'] == 0:
+                return True
+            if name == 'LegacyCSI' and raw[name]['Info']['DeviceType']  == 0:
+                return True
+            if name == 'BasebandSignals' and raw[name]['Info']['Ndim'] == 0:
+                return True
+            if name == 'PreEQSymbols' and raw[name]['Info']['Ndim'] == 0:
+                return True
+            if name == 'MPDU' and raw[name]['Info']['Length'] == 0:
+                return True
+            return False
+
+        def report(s, raw, parent=None, indent=0):
+            indent += 2
+            tab = " " * indent
+            names = rfn.get_names(raw.dtype)
+            for i, name in enumerate(names):
+                if isinstance(name, tuple):
+                    if skip(raw, name[0]):
+                        continue
+                    s += T % (tab, name[0], '')
+                    s = report(s, raw[name[0]], name[0], indent)
+                else:
+                    if parent and parent.endswith("ExtraInfo"):
+                        if name.startswith("Has") or not raw[names[i - hfo]]:
+                            continue
+                    if name == 'Shape':
+                        s += T % (tab, name, tuple(raw[name]))
+                    elif name == 'Majority':
+                        s += T % (tab, name, raw[name].tobytes())
+                    elif name.lower() == "devicetype":
+                        s += T % (tab, name, hex(raw[name]))
+                    elif name.lower() == "tuning_policy":
+                        s += T % (tab, name, AtherosCFTuningPolicy[raw[name]])
+                    elif raw[name].size == 6:
+                        s += T % (tab, name, stringify(raw[name]))
+                    elif raw[name].size > 10:
+                        s += T % (tab, name, raw[name].shape)
+                    else:
+                        s += T % (tab, name, raw[name])
+            return s
+
+        s = "%dth frame:\n" % index
+        s += T % ("  ", "file", self.file)
+        s += T % ("  ", "count", self.count)
+        s = report(s, self.raw[index], None, 0)
+        print(s, end='')
